@@ -1,4 +1,7 @@
+# -*- encoding: utf-8 -*-
+#
 # Copyright (c) 2013 Paul Tagliamonte <paultag@debian.org>
+# Copyright (c) 2013 Julien Danjou <julien@danjou.info>
 #
 # Permission is hereby granted, free of charge, to any person obtaining a
 # copy of this software and associated documentation files (the "Software"),
@@ -29,6 +32,7 @@ from hy.models.dict import HyDict
 
 from hy.util import flatten_literal_list
 
+import codecs
 import ast
 import sys
 
@@ -40,6 +44,20 @@ class HyCompileError(HyError):
 _compile_table = {}
 
 
+def ast_str(foobar):
+    if sys.version_info[0] >= 3:
+        return str(foobar)
+
+    try:
+        return str(foobar)
+    except UnicodeEncodeError:
+        pass
+
+    enc = codecs.getencoder('punycode')
+    foobar, _ = enc(foobar)
+    return "__hy_%s" % (str(foobar).replace("-", "_"))
+
+
 def builds(_type):
     def _dec(fn):
         _compile_table[_type] = fn
@@ -50,6 +68,38 @@ def builds(_type):
     return _dec
 
 
+def _raise_wrong_args_number(expression, error):
+    err = TypeError(error % (expression.pop(0),
+                             len(expression)))
+    err.start_line = expression.start_line
+    err.start_column = expression.start_column
+    raise err
+
+
+def checkargs(exact=None, min=None, max=None):
+    def _dec(fn):
+        def checker(self, expression):
+            if exact is not None and (len(expression) - 1) != exact:
+                _raise_wrong_args_number(expression,
+                                         "`%%s' needs %d arguments, got %%d" %
+                                         exact)
+
+            if min is not None and (len(expression) - 1) < min:
+                _raise_wrong_args_number(
+                    expression,
+                    "`%%s' needs at least %d arguments, got %%d" % (min))
+
+            if max is not None and (len(expression) - 1) > max:
+                _raise_wrong_args_number(
+                    expression,
+                    "`%%s' needs at most %d arguments, got %%d" % (max))
+
+            return fn(self, expression)
+
+        return checker
+    return _dec
+
+
 class HyASTCompiler(object):
 
     def __init__(self):
@@ -57,9 +107,16 @@ class HyASTCompiler(object):
         self.anon_fn_count = 0
 
     def compile(self, tree):
-        for _type in _compile_table:
-            if type(tree) == _type:
-                return _compile_table[_type](self, tree)
+        try:
+            for _type in _compile_table:
+                if type(tree) == _type:
+                    return _compile_table[_type](self, tree)
+        except Exception as e:
+            err = HyCompileError(str(e))
+            err.exception = e
+            err.start_line = getattr(e, "start_line", None)
+            err.start_column = getattr(e, "start_column", None)
+            raise err
 
         raise HyCompileError("Unknown type - `%s'" % (str(type(tree))))
 
@@ -100,10 +157,13 @@ class HyASTCompiler(object):
 
     @builds("do")
     @builds("progn")
+    @checkargs(min=1)
     def compile_do_expression(self, expr):
         return [self.compile(x) for x in expr[1:]]
 
     @builds("throw")
+    @builds("raise")
+    @checkargs(min=1)
     def compile_throw_expression(self, expr):
         expr.pop(0)
         exc = self.compile(expr.pop(0))
@@ -125,37 +185,108 @@ class HyASTCompiler(object):
         else:
             Try = ast.TryExcept
 
+        if len(expr) == 0:
+            # (try)
+            body = [ast.Pass(lineno=expr.start_line,
+                             col_offset=expr.start_column)]
+        else:
+            # (try something…)
+            body = self._code_branch(self.compile(expr.pop(0)))
+
+        if len(expr) == 0:
+            # (try) or (try body)
+            handlers = [ast.ExceptHandler(
+                lineno=expr.start_line,
+                col_offset=expr.start_column,
+                type=None,
+                name=None,
+                body=[ast.Pass(lineno=expr.start_line,
+                               col_offset=expr.start_column)])]
+        else:
+            # (try body except except…)
+            handlers = [self.compile(s) for s in expr]
+
         return Try(
             lineno=expr.start_line,
             col_offset=expr.start_column,
-            body=self._code_branch(self.compile(expr.pop(0))),
-            handlers=[self.compile(s) for s in expr],
+            body=body,
+            handlers=handlers,
             finalbody=[],
             orelse=[])
 
     @builds("catch")
+    @builds("except")
     def compile_catch_expression(self, expr):
         expr.pop(0)  # catch
-        _type = self.compile(expr.pop(0))
-        name = expr.pop(0)
 
-        if sys.version_info[0] >= 3:
-            # Python3 features a change where the Exception handler
-            # moved the name from a Name() to a pure Python String type.
-            #
-            # We'll just make sure it's a pure "string", and let it work
-            # it's magic.
-            name = str(name)
+        try:
+            exceptions = expr.pop(0)
+        except IndexError:
+            exceptions = []
+        # exceptions catch should be either:
+        # [[list of exceptions]]
+        # or
+        # [variable [list of exceptions]]
+        # or
+        # [variable exception]
+        # or
+        # [exception]
+        # or
+        # []
+        if len(exceptions) > 2:
+            raise TypeError("`catch' exceptions list is too long")
+
+        # [variable [list of exceptions]]
+        # let's pop variable and use it as name
+        if len(exceptions) == 2:
+            name = exceptions.pop(0)
+            if sys.version_info[0] >= 3:
+                # Python3 features a change where the Exception handler
+                # moved the name from a Name() to a pure Python String type.
+                #
+                # We'll just make sure it's a pure "string", and let it work
+                # it's magic.
+                name = ast_str(name)
+            else:
+                # Python2 requires an ast.Name, set to ctx Store.
+                name = self._storeize(self.compile(name))
         else:
-            # Python2 requires an ast.Name, set to ctx Store.
-            name = self._storeize(self.compile(name))
+            name = None
+
+        try:
+            exceptions_list = exceptions.pop(0)
+        except IndexError:
+            exceptions_list = []
+
+        if isinstance(exceptions_list, list):
+            if len(exceptions_list):
+                # [FooBar BarFoo] → catch Foobar and BarFoo exceptions
+                _type = ast.Tuple(elts=[self.compile(x)
+                                        for x in exceptions_list],
+                                  lineno=expr.start_line,
+                                  col_offset=expr.start_column,
+                                  ctx=ast.Load())
+            else:
+                # [] → all exceptions catched
+                _type = None
+        elif isinstance(exceptions_list, HySymbol):
+            _type = self.compile(exceptions_list)
+        else:
+            raise TypeError("`catch' needs a valid exception list to catch")
+
+        if len(expr) == 0:
+            # No body
+            body = [ast.Pass(lineno=expr.start_line,
+                             col_offset=expr.start_column)]
+        else:
+            body = self._code_branch([self.compile(x) for x in expr])
 
         return ast.ExceptHandler(
             lineno=expr.start_line,
             col_offset=expr.start_column,
             type=_type,
             name=name,
-            body=self._code_branch([self.compile(x) for x in expr]))
+            body=body)
 
     def _code_branch(self, branch):
         if isinstance(branch, list):
@@ -163,21 +294,16 @@ class HyASTCompiler(object):
         return self._mangle_branch([branch])
 
     @builds("if")
+    @checkargs(min=2, max=3)
     def compile_if_expression(self, expr):
-        expr.pop(0)
-        try:
-            test = expr.pop(0)
-        except IndexError:
-            raise TypeError("if expects at least 2 arguments, got 0")
-        test = self.compile(test)
-        try:
-            body = expr.pop(0)
-        except IndexError:
-            raise TypeError("if expects at least 2 arguments, got 1")
-        body = self._code_branch(self.compile(body))
-        orel = []
-        if len(expr) > 0:
+        expr.pop(0)             # if
+        test = self.compile(expr.pop(0))
+        body = self._code_branch(self.compile(expr.pop(0)))
+
+        if len(expr) == 1:
             orel = self._code_branch(self.compile(expr.pop(0)))
+        else:
+            orel = []
 
         return ast.If(test=test,
                       body=body,
@@ -206,6 +332,7 @@ class HyASTCompiler(object):
             nl=True)
 
     @builds("assert")
+    @checkargs(1)
     def compile_assert_expression(self, expr):
         expr.pop(0)  # assert
         e = expr.pop(0)
@@ -215,6 +342,7 @@ class HyASTCompiler(object):
                           col_offset=e.start_column)
 
     @builds("lambda")
+    @checkargs(min=2)
     def compile_lambda_expression(self, expr):
         expr.pop(0)
         sig = expr.pop(0)
@@ -224,7 +352,7 @@ class HyASTCompiler(object):
             lineno=expr.start_line,
             col_offset=expr.start_column,
             args=ast.arguments(args=[
-                ast.Name(arg=str(x), id=str(x),
+                ast.Name(arg=ast_str(x), id=ast_str(x),
                          ctx=ast.Param(),
                          lineno=x.start_line,
                          col_offset=x.start_column)
@@ -237,10 +365,12 @@ class HyASTCompiler(object):
             body=self.compile(body))
 
     @builds("pass")
+    @checkargs(0)
     def compile_pass_expression(self, expr):
         return ast.Pass(lineno=expr.start_line, col_offset=expr.start_column)
 
     @builds("yield")
+    @checkargs(1)
     def compile_yield_expression(self, expr):
         expr.pop(0)
         return ast.Yield(
@@ -254,7 +384,7 @@ class HyASTCompiler(object):
         return ast.Import(
             lineno=expr.start_line,
             col_offset=expr.start_column,
-            names=[ast.alias(name=str(x), asname=None) for x in expr])
+            names=[ast.alias(name=ast_str(x), asname=None) for x in expr])
 
     @builds("import_as")
     def compile_import_as_expression(self, expr):
@@ -263,21 +393,23 @@ class HyASTCompiler(object):
         return ast.Import(
             lineno=expr.start_line,
             col_offset=expr.start_column,
-            module=str(expr.pop(0)),
-            names=[ast.alias(name=str(x[0]),
-                             asname=str(x[1])) for x in modlist])
+            module=ast_str(expr.pop(0)),
+            names=[ast.alias(name=ast_str(x[0]),
+                             asname=ast_str(x[1])) for x in modlist])
 
     @builds("import_from")
+    @checkargs(min=1)
     def compile_import_from_expression(self, expr):
         expr.pop(0)  # index
         return ast.ImportFrom(
             lineno=expr.start_line,
             col_offset=expr.start_column,
-            module=str(expr.pop(0)),
-            names=[ast.alias(name=str(x), asname=None) for x in expr],
+            module=ast_str(expr.pop(0)),
+            names=[ast.alias(name=ast_str(x), asname=None) for x in expr],
             level=0)
 
     @builds("get")
+    @checkargs(2)
     def compile_index_expression(self, expr):
         expr.pop(0)  # index
         val = self.compile(expr.pop(0))  # target
@@ -291,6 +423,7 @@ class HyASTCompiler(object):
             ctx=ast.Load())
 
     @builds("slice")
+    @checkargs(min=1, max=3)
     def compile_slice_expression(self, expr):
         expr.pop(0)  # index
         val = self.compile(expr.pop(0))  # target
@@ -313,6 +446,7 @@ class HyASTCompiler(object):
             ctx=ast.Load())
 
     @builds("assoc")
+    @checkargs(3)
     def compile_assoc_expression(self, expr):
         expr.pop(0)  # assoc
         # (assoc foo bar baz)  => foo[bar] = baz
@@ -333,6 +467,7 @@ class HyASTCompiler(object):
             value=self.compile(val))
 
     @builds("decorate_with")
+    @checkargs(min=1)
     def compile_decorate_expression(self, expr):
         expr.pop(0)  # decorate-with
         fn = self.compile(expr.pop(-1))
@@ -341,11 +476,21 @@ class HyASTCompiler(object):
         fn.decorator_list = [self.compile(x) for x in expr]
         return fn
 
-    @builds("with_as")
-    def compile_with_as_expression(self, expr):
-        expr.pop(0)  # with-as
-        ctx = self.compile(expr.pop(0))
-        thing = self._storeize(self.compile(expr.pop(0)))
+    @builds("with")
+    @checkargs(min=2)
+    def compile_with_expression(self, expr):
+        expr.pop(0)  # with
+
+        args = expr.pop(0)
+        if len(args) > 2 or len(args) < 1:
+            raise TypeError("with needs [arg (expr)] or [(expr)]")
+
+        args.reverse()
+        ctx = self.compile(args.pop(0))
+
+        thing = None
+        if args != []:
+            thing = self._storeize(self.compile(args.pop(0)))
 
         ret = ast.With(context_expr=ctx,
                        lineno=expr.start_line,
@@ -368,6 +513,7 @@ class HyASTCompiler(object):
                          ctx=ast.Load())
 
     @builds("list_comp")
+    @checkargs(min=2, max=3)
     def compile_list_comprehension(self, expr):
         # (list-comp expr (target iter) cond?)
         expr.pop(0)
@@ -402,6 +548,7 @@ class HyASTCompiler(object):
         return name
 
     @builds("kwapply")
+    @checkargs(2)
     def compile_kwapply_expression(self, expr):
         expr.pop(0)  # kwapply
         call = self.compile(expr.pop(0))
@@ -410,11 +557,39 @@ class HyASTCompiler(object):
         if type(call) != ast.Call:
             raise TypeError("kwapplying a non-call")
 
-        call.keywords = [ast.keyword(arg=str(x),
+        call.keywords = [ast.keyword(arg=ast_str(x),
                          value=self.compile(kwargs[x])) for x in kwargs]
 
         return call
 
+    @builds("not")
+    @builds("~")
+    @checkargs(1)
+    def compile_unary_operator(self, expression):
+        ops = {"not": ast.Not,
+               "~": ast.Invert}
+        operator = expression.pop(0)
+        operand = expression.pop(0)
+        return ast.UnaryOp(op=ops[operator](),
+                           operand=self.compile(operand),
+                           lineno=operator.start_line,
+                           col_offset=operator.start_column)
+
+    @builds("and")
+    @builds("or")
+    @checkargs(min=2)
+    def compile_logical_or_and_and_operator(self, expression):
+        ops = {"and": ast.And,
+               "or": ast.Or}
+        operator = expression.pop(0)
+        values = []
+        for child in expression:
+            values.append(self.compile(child))
+        return ast.BoolOp(op=ops[operator](),
+                          lineno=operator.start_line,
+                          col_offset=operator.start_column,
+                          values=values)
+
     @builds("=")
     @builds("!=")
     @builds("<")
@@ -425,6 +600,7 @@ class HyASTCompiler(object):
     @builds("in")
     @builds("is_not")
     @builds("not_in")
+    @checkargs(min=2)
     def compile_compare_op_expression(self, expression):
         ops = {"=": ast.Eq, "!=": ast.NotEq,
                "<": ast.Lt, "<=": ast.LtE,
@@ -448,6 +624,7 @@ class HyASTCompiler(object):
     @builds("-")
     @builds("/")
     @builds("*")
+    @checkargs(min=2)
     def compile_maths_expression(self, expression):
         # operator = Mod | Pow | LShift | RShift | BitOr |
         #            BitXor | BitAnd | FloorDiv
@@ -486,7 +663,7 @@ class HyASTCompiler(object):
                 lineno=expr.start_line,
                 col_offset=expr.start_column,
                 value=self.compile(obj),
-                attr=str(fn),
+                attr=ast_str(fn),
                 ctx=ast.Load()),
             args=[self.compile(x) for x in expr],
             keywords=[],
@@ -516,6 +693,7 @@ class HyASTCompiler(object):
     @builds("def")
     @builds("setf")
     @builds("setv")
+    @checkargs(2)
     def compile_def_expression(self, expression):
         expression.pop(0)  # "def"
         name = expression.pop(0)
@@ -526,7 +704,7 @@ class HyASTCompiler(object):
             # We special case a FunctionDef, since we can define by setting
             # FunctionDef's .name attribute, rather then foo == anon_fn. This
             # helps keep things clean.
-            what.name = str(name)
+            what.name = ast_str(name)
             return what
 
         name = self._storeize(self.compile(name))
@@ -537,6 +715,7 @@ class HyASTCompiler(object):
             targets=[name], value=what)
 
     @builds("foreach")
+    @checkargs(min=1)
     def compile_for_expression(self, expression):
         ret_status = self.returnable
         self.returnable = False
@@ -544,9 +723,6 @@ class HyASTCompiler(object):
         expression.pop(0)  # for
         name, iterable = expression.pop(0)
         target = self._storeize(self.compile_symbol(name))
-        # support stuff like:
-        # (for [x [1 2 3 4]
-        #       y [a b c d]] ...)
 
         ret = ast.For(lineno=expression.start_line,
                       col_offset=expression.start_column,
@@ -560,17 +736,10 @@ class HyASTCompiler(object):
         return ret
 
     @builds("while")
+    @checkargs(min=2)
     def compile_while_expression(self, expr):
         expr.pop(0)  # "while"
-
-        try:
-            test = expr.pop(0)
-        except IndexError:
-            raise TypeError("while expects at least 2 arguments, got 0")
-        test = self.compile(test)
-
-        if not expr:
-            raise TypeError("while expects a body")
+        test = self.compile(expr.pop(0))
 
         return ast.While(test=test,
                          body=self._mangle_branch([
@@ -588,16 +757,32 @@ class HyASTCompiler(object):
             col_offset=expr.start_column)
 
     @builds("fn")
+    @checkargs(min=2)
     def compile_fn_expression(self, expression):
         expression.pop(0)  # fn
 
         ret_status = self.returnable
-        self.returnable = True
 
         self.anon_fn_count += 1
         name = "_hy_anon_fn_%d" % (self.anon_fn_count)
         sig = expression.pop(0)
 
+        body = []
+        if expression != []:
+            self.returnable = True
+            tailop = self.compile(expression.pop(-1))
+            self.returnable = False
+            for el in expression:
+                body.append(self.compile(el))
+            body.append(tailop)
+
+        if body == []:
+            body = [ast.Pass(lineno=expression.start_line,
+                             col_offset=expression.start_column)]
+
+        self.returnable = True
+        body = self._code_branch(body)
+
         ret = ast.FunctionDef(
             name=name,
             lineno=expression.start_line,
@@ -605,7 +790,7 @@ class HyASTCompiler(object):
             args=ast.arguments(
                 args=[
                     ast.Name(
-                        arg=str(x), id=str(x),
+                        arg=ast_str(x), id=ast_str(x),
                         ctx=ast.Param(),
                         lineno=x.start_line,
                         col_offset=x.start_column)
@@ -615,8 +800,7 @@ class HyASTCompiler(object):
                 kwonlyargs=[],
                 kw_defaults=[],
                 defaults=[]),
-            body=self._code_branch([
-                self.compile(x) for x in expression]),
+            body=body,
             decorator_list=[])
 
         self.returnable = ret_status
@@ -639,19 +823,19 @@ class HyASTCompiler(object):
                 lineno=symbol.start_line,
                 col_offset=symbol.start_column,
                 value=self.compile_symbol(glob),
-                attr=str(local),
+                attr=ast_str(local),
                 ctx=ast.Load()
             )
 
-        return ast.Name(id=str(symbol),
-                        arg=str(symbol),
+        return ast.Name(id=ast_str(symbol),
+                        arg=ast_str(symbol),
                         ctx=ast.Load(),
                         lineno=symbol.start_line,
                         col_offset=symbol.start_column)
 
     @builds(HyString)
     def compile_string(self, string):
-        return ast.Str(s=str(string), lineno=string.start_line,
+        return ast.Str(s=ast_str(string), lineno=string.start_line,
                        col_offset=string.start_column)
 
     @builds(HyDict)
