@@ -1,4 +1,4 @@
-# Copyright (c) 2012 Paul Tagliamonte <paultag@debian.org>
+# Copyright (c) 2013 Paul Tagliamonte <paultag@debian.org>
 #
 # Permission is hereby granted, free of charge, to any person obtaining a
 # copy of this software and associated documentation files (the "Software"),
@@ -31,3 +31,6 @@ class HyList(HyObject, list):
             x.replace(other)
 
         HyObject.replace(self, other)
+
+    def __repr__(self):
+        return "[%s]" % (" ".join([str(x) for x in self]))
