@@ -23,18 +23,15 @@ import sys
 
 
 if sys.version_info[0] >= 3:
-    _str_type = str
+    str_type = str
 else:
-    _str_type = unicode
+    str_type = unicode
 
 
-class HyString(HyObject, _str_type):
+class HyString(HyObject, str_type):
     """
     Generic Hy String object. Helpful to store string literals from Hy
     scripts. It's either a ``str`` or a ``unicode``, depending on the
     Python version.
     """
-
-    def __new__(cls, value):
-        obj = _str_type.__new__(cls, value)
-        return obj
+    pass
